/*******************************************************************************
 *******************************************************************************
 **                                                                           **
 **                                                                           **
 **  Copyright 2003-2020 Tom Bascom, Greenfield Technologies                  **
 **  http://www.greenfieldtech.com                                            **
 **                                                                           **
 **  ProTop is free software; you can redistribute it and/or modify it        **
 **  under the terms of the GNU General Public License (GPL) as published     **
 **  by the Free Software Foundation; either version 2 of the License, or     **
 **  at your option) any later version.                                       **
 **                                                                           **
 **  ProTop is distributed in the hope that it will be useful, but WITHOUT    **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    **
 **  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License     **
 **  for more details.                                                        **
 **                                                                           **
 **  See TERMS.TXT for more information regarding the Terms and Conditions    **
 **  of use and alternative licensing options for this software.              **
 **                                                                           **
 **  A copy of the GPL is in GPL.TXT which was provided with this package.    **
 **                                                                           **
 **  See http://www.fsf.org for more information about the GPL.               **
 **                                                                           **
 **                                                                           **
 *******************************************************************************
 *******************************************************************************
 *
 * lobact.p
 *
 *
 * LOB Activity
 *
 *
 * Author:
 *
 *	Tom Bascom, Greenfield Technologies
 *	http://www.greenfieldtech.com
 *	September 5, 2003
 *
 */

{lib/protop.i}
{lib/tick.i}

define output parameter dcDescription as character no-undo initial "LOBActivity".

{lib/tt_lobdata.i}

define temp-table tt_LOBAct no-undo
  field xid        as integer   format "->>>>9"
  field LOBNum     as integer   format "->>>>9"          label "LOB#"
  field LOBAreaNum as integer   format ">>>>9"           label "Area#"
  field LOBPool    as character format "x(3)"            label "BX"
  field LOBName    as character format "x(45)"           label "LOB Name"
  field LOBRecs    as decimal   format ">>>>>>>>>9"      label "#Recs"
  field LOBMinRow  as decimal   format ">>>>>>>>>9"      label "Min Sz"
  field LOBMaxRow  as decimal   format ">>>>>>>>>>9"     label "Max Sz"
  field LOBAvgRow  as decimal   format ">>>>>>>>>>9"     label "Avg Sz"
  field LOBCr      as decimal   format ">>>>>>>>9"       label "Create"
  field LOBRd      as decimal   format ">>>>>>>>>9"      label "Read"
  field LOBUp      as decimal   format ">>>>>>>>9"       label "Update"
  field LOBDl      as decimal   format ">>>>>>>>9"       label "Delete"
  field LOBOSRd    as decimal   format ">>>>>>>>9"       label "OS Read"

  index xid-idx  is unique xid
  index LOBRd-idx  is primary LOBRd descending
.

define new global shared variable rowLimit as integer no-undo.


procedure mon-init:

  define variable b as handle.

  b = ( temp-table tt_LOBAct:default-buffer-handle ).

  if pt_altWidth = "80" then
    do:
      altFormat( "80", b, "LOBName",  "x(30)" ).
      altFormat( "80", b, "LOBAvgRow",">>>>>>>>9" ).
      altFormat( "80", b, "LOBCr",    ">>>>>>>>9" ).
      altFormat( "80", b, "LOBRd",    ">>>>>>>>9" ).
      altFormat( "80", b, "LOBUp",    ">>>>>>>>9" ).
      altFormat( "80", b, "LOBDl",    ">>>>>>>>9" ).
    end.

  run updTick.

  return.

end.


procedure mon-update:

  define input parameter argList as character no-undo.

  define variable i        as integer no-undo.
  define variable topLimit as integer no-undo initial 50.				/* formerly 10				*/

  run updTick.

&IF DECIMAL(SUBSTRING(PROVERSION,1,INDEX(PROVERSION,".") + 1)) >= 12.0 &THEN

  for each tt_lob:
    run upd-tt_LOBdata ( tt_lob.xid ).
  end.

  empty temp-table tt_LOBAct.

  /* top X reads
   */

  i = 0.
  for each tt_LOBdata no-lock by tt_LOBdata.LOB-rd[x] descending:

    find tt_LOBAct where tt_LOBAct.xid = tt_LOBdata.xid no-error.			/* prevent duplicates			*/
    if available tt_LOBAct then next.

    create tt_LOBAct.
    assign
      tt_LOBAct.xid        = tt_LOBdata.xid
      tt_LOBAct.LOBNum     = tt_LOBdata.xid
      tt_LOBAct.LOBAreaNum = tt_LOBdata.areaNum
      tt_LOBAct.LOBPool    = tt_LOBdata.LOBPool
  /*  tt_LOBAct.tblNm      = tt_LOBdata.tblNm   */
  /*  tt_LOBAct.LOBName    = tt_LOBdata.LOBName */
      tt_LOBAct.LOBName    = tt_LOBdata.tblNm + "." + tt_LOBdata.LOBName
      tt_LOBAct.LOBCr      = ( tt_LOBdata.LOB-cre[x]  / z )
      tt_LOBAct.LOBRd      = ( tt_LOBdata.LOB-rd[x]   / z )
      tt_LOBAct.LOBUp      = ( tt_LOBdata.LOB-upd[x]  / z )
      tt_LOBAct.LOBDl      = ( tt_LOBdata.LOB-del[x]  / z )
      tt_LOBAct.LOBOSRd    = ( tt_LOBdata.LOB-osrd[x] / z )
    .

    run getLOBInfo(
      tt_LOBAct.LOBName,
      output tt_LOBAct.LOBRecs,
      output tt_LOBAct.LOBMinRow,
      output tt_LOBAct.LOBMaxRow,
      output tt_LOBAct.LOBAvgRow
    ).

  end.

&ENDIF

  add2ds( temp-table tt_LOBAct:default-buffer-handle ).

  return.

end.

return.
